from rest_framework import viewsets
from knox.auth import TokenAuthentication
from rest_framework.permissions import IsAdminUser, IsAuthenticated
from rest_framework import serializers as slzrs
from rest_framework.serializers import HyperlinkedIdentityField
from django.contrib.auth.models import AnonymousUser

from jaseci_serv.base import models
from jaseci_serv.base.orm_hook import json_str_to_jsci_dict


class JaseciObjectSerializer(slzrs.HyperlinkedModelSerializer):
    """Serializer for Jaseci object model"""
    url = HyperlinkedIdentityField(view_name="obj_api:jaseciobject-detail")

    class Meta:
        model = models.JaseciObject
        fields = (
            'jid', 'j_parent', 'j_master', 'j_access', 'j_r_acc_ids',
            'j_rw_acc_ids', 'j_type', 'url',
            'j_type', 'name', 'kind', 'j_timestamp', 'jsci_obj'
        )
        read_only_Fields = ('id', 'j_type', 'timestamp')

    def to_representation(self, instance):
        """Convert jsci_obj to dictionary so entire payload is one JSON"""
        ret = super(JaseciObjectSerializer, self).to_representation(instance)
        if(len(ret['jsci_obj'])):
            ret['jsci_obj'] = json_str_to_jsci_dict(ret['jsci_obj'])
        return ret


class ObjectViewSet(viewsets.ModelViewSet):
    """Edit Jaseci object through the api"""
    authentication_classes = (TokenAuthentication,)
    permission_classes = (IsAuthenticated,)
    queryset = models.JaseciObject.objects.all()
    serializer_class = JaseciObjectSerializer

    def get_queryset(self):
        """Return objects for the current authenticated user only"""
        if (isinstance(self.request.user, AnonymousUser)):
            return None
        else:
            return self.queryset.filter(
                j_master=self.request.user.master.urn
            ).order_by('-name')

    def perform_create(self, serializer):
        """Create a new object"""
        serializer.save(j_master=self.request.user.master.urn)


class GlobalVarsSerializer(slzrs.ModelSerializer):
    """Serializer for Global Config model"""

    class Meta:
        model = models.GlobalVars
        fields = (
            'name', 'value'
        )


class ConfigViewSet(viewsets.ModelViewSet):
    """Edit Global Config through the api"""
    authentication_classes = (TokenAuthentication,)
    permission_classes = (IsAuthenticated, IsAdminUser)
    queryset = models.GlobalVars.objects.all()
    serializer_class = GlobalVarsSerializer

    def get_queryset(self):
        """Return objects for the current authenticated user only"""
        return self.queryset.all().order_by('-name')



class mock_redis:
    def __init__(self, cache=dict()):
        self.cache = cache

    def get(self, key):
        if key in self.cache:
            return self.cache[key]
        return None  # return nil

    def set(self, key, value, *args, **kwargs):
        self.cache[key] = value
        return "OK"

    def hget(self, hash, key):
        if hash in self.cache:
            if key in self.cache[hash]:
                return self.cache[hash][key]
        return None  # return nil

    def hset(self, hash, key, value, *args, **kwargs):
        self.cache[hash][key] = value
        return 1

    def exists(self, key):
        if key in self.cache:
            return 1
        return 0

    def delete(self, key):
        del self.cache[key]
        return None  # return nil

    def cache_overwrite(self, cache=dict()):
        self.cache = cache



from django.contrib.auth import get_user_model
from django.urls import reverse
from rest_framework.test import APIClient
from jaseci.utils.utils import TestCaseHelper
import jaseci.actions.live_actions as lact
from django.test import TestCase
import uuid
import base64
import datetime
import random
import os


class test_ll(TestCaseHelper, TestCase):
    """Test the authorized user node API"""

    def setUp(self):
        super().setUp()
        lact.load_local_actions(os.path.dirname(__file__) + '/infer.py')
        self.user = get_user_model().objects.create_user(
            'JSCITfdfdEST_test@jaseci.com',
            'password'
        )
        self.client = APIClient()
        self.client.force_authenticate(self.user)
        ll_loc = os.path.dirname(__file__) + '/ll.jac'
        ll_file = base64.b64encode(
            open(ll_loc).read().encode()).decode()
        payload = {'op': 'sentinel_register',
                   'name': 'Something', 'code': ll_file, 'encoded': True}
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        payload = {'op': 'global_sentinel_set'}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format='json')
        self.user = get_user_model().objects.create_user(
            'J2SCITfdfdEST_test@jaseci.com',
            'password'
        )
        self.client = APIClient()
        self.client.force_authenticate(self.user)
        self.master = self.user.get_master()
        payload = {'op': 'sentinel_active_global'}
        res = self.client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format='json')
        payload = {'op': 'graph_create'}
        self.snt = self.master._h.get_obj(
            self.master.jid, uuid.UUID(res.data['sentinel']['jid']))
        res = self.client.post(reverse(f'jac_api:{payload["op"]}'), payload)
        self.gph = self.master._h.get_obj(
            self.master.jid, uuid.UUID(res.data['jid']))

    def tearDown(self):
        super().tearDown()

    def run_walker(self, w_name, ctx, prime=None):
        """Helper to make calls to execute walkers"""
        if(not prime):
            payload = {'snt': self.snt.id.urn, 'name': w_name,
                       'nd': self.gph.id.urn, 'ctx': ctx}
        else:
            payload = {'snt': self.snt.id.urn, 'name': w_name,
                       'nd': prime, 'ctx': ctx}
        res = self.client.post(
            reverse('jac_api:walker_run'), payload, format='json')
        return res.data

    def graph_node_set(self, nd_id, ctx):
        """Helper to set node context"""
        payload = {'snt': self.snt.id.urn, 'nd': nd_id, 'ctx': ctx}
        self.client.post(
            reverse('jac_api:graph_node_set'), payload, format='json')

    def test_ll_today_new(self):
        """Test LifeLogify Jac Implementation"""
        data = self.run_walker('get_gen_day', {})['report']
        self.assertEqual(data[0][1]['name'], 'day')
        jid = data[0][1]['jid']
        data = self.run_walker('get_gen_day', {})['report']
        self.assertEqual(data[0][1]['jid'], jid)

    def test_ll_create_get_workette(self):
        """Test LifeLogify Jac Implementation"""
        data = self.run_walker('get_gen_day', {})['report']
        self.assertEqual(data[0][1]['name'], 'day')
        jid = data[0][1]['jid']
        data = self.run_walker('create_workette', {}, prime=jid)['report']
        self.assertEqual(data[0]['name'], 'workette')
        data = self.run_walker('create_workette', {}, prime=jid)['report']
        self.assertEqual(data[0]['name'], 'workette')
        data = self.run_walker('get_workettes', {}, prime=jid)['report']
        self.assertEqual(len(data), 2)
        self.assertEqual(data[0]['name'], 'workette')
        self.assertEqual(data[1]['name'], 'workette')
        wjid = data[0]['jid']
        data = self.run_walker('create_workette', {}, prime=wjid)['report']
        data = self.run_walker('create_workette', {}, prime=wjid)['report']
        data = self.run_walker('create_workette', {}, prime=wjid)['report']
        data = self.run_walker('get_workettes', {}, prime=wjid)['report']
        self.assertEqual(len(data), 3)
        data = self.run_walker('get_workettes', {}, prime=jid)['report']
        self.assertEqual(len(data), 2)
        data = self.run_walker('get_workettes_deep', {}, prime=jid)['report']
        self.assertEqual(len(data), 5)

    def test_ll_get_date(self):
        """Test LifeLogify Jac Implementation"""
        data = self.run_walker('get_gen_day', {})
        data = self.run_walker('gen_rand_life', {})
        data = self.run_walker('get_latest_day', {'show_report': 1})['report']
        wjid = data[0][1]['jid']
        data = self.run_walker('get_workettes', {}, prime=wjid)['report']
        self.assertEqual(len(data), 0)
        data = self.run_walker('get_latest_day', {'before_date': '2020-03-01',
                                                  'show_report': 1})['report']
        wjid = data[0][1]['jid']
        data = self.run_walker('get_workettes', {}, prime=wjid)['report']
        self.assertGreater(len(data), 1)

    def test_ll_carry_forward_simple(self):
        """Test LifeLogify Jac Implementation"""
        data = self.run_walker('gen_rand_life', {})
        data = self.run_walker('get_gen_day', {})
        wjid = data['report'][0][1]['jid']
        data = self.run_walker('get_workettes', {}, prime=wjid)
        self.assertGreater(len(data['report']), 1)

    def test_ll_delete_workette(self):
        """Test LifeLogify Jac Implementation"""
        data = self.run_walker('get_gen_day', {})['report']
        self.assertEqual(data[0][1]['name'], 'day')
        jid = data[0][1]['jid']
        data = self.run_walker('create_workette', {}, prime=jid)['report']
        self.assertEqual(data[0]['name'], 'workette')
        data = self.run_walker('create_workette', {}, prime=jid)['report']
        self.assertEqual(data[0]['name'], 'workette')
        data = self.run_walker('get_workettes', {}, prime=jid)['report']
        self.assertEqual(len(data), 2)
        self.assertEqual(data[0]['name'], 'workette')
        self.assertEqual(data[1]['name'], 'workette')
        wjid = data[0]['jid']
        data = self.run_walker('create_workette', {}, prime=wjid)
        data = self.run_walker('create_workette', {}, prime=wjid)
        data = self.run_walker('create_workette', {}, prime=wjid)
        data = self.run_walker('get_workettes', {}, prime=wjid)['report']
        self.assertEqual(len(data), 3)
        data = self.run_walker('delete_workette', {}, prime=wjid)
        data = self.run_walker('get_workettes', {}, prime=jid)['report']
        self.assertEqual(len(data), 1)
        data = self.run_walker('get_workettes_deep', {}, prime=jid)['report']
        self.assertEqual(len(data), 1)

    def test_ll_delete_workette_no_leaks(self):
        """Test LifeLogify Jac Implementation"""
        data = self.run_walker('get_gen_day', {})['report']
        self.assertEqual(data[0][1]['name'], 'day')
        jid = data[0][1]['jid']
        data = self.run_walker('create_workette', {}, prime=jid)['report']
        self.assertEqual(data[0]['name'], 'workette')
        len_before = self.master._h.get_object_distribution()
        data = self.run_walker('create_workette', {}, prime=jid)['report']
        self.assertEqual(data[0]['name'], 'workette')
        data = self.run_walker('get_workettes', {}, prime=jid)['report']
        self.assertEqual(len(data), 2)
        self.assertEqual(data[0]['name'], 'workette')
        self.assertEqual(data[1]['name'], 'workette')
        wjid = data[1]['jid']
        data = self.run_walker('create_workette', {}, prime=wjid)['report']
        exjid = data[0]['jid']
        data = self.run_walker('create_workette', {}, prime=wjid)
        data = self.run_walker('create_workette', {}, prime=wjid)
        data = self.run_walker('get_workettes', {}, prime=wjid)['report']
        self.assertEqual(len(data), 3)
        data = self.run_walker('delete_workette', {}, prime=wjid)['report']
        self.assertNotIn(uuid.UUID(exjid), self.master._h.mem.keys())
        self.assertIn(uuid.UUID(jid), self.master._h.mem.keys())
        data = self.run_walker('get_workettes', {}, prime=jid)['report']
        self.assertEqual(len(data), 1)
        data = self.run_walker('get_workettes_deep', {}, prime=jid)['report']
        self.assertEqual(len(data), 1)
        len_after = self.master._h.get_object_distribution()
        self.assertEqual(len_before, len_after)

    def test_parent_suggestion(self):
        if (not lact.load_remote_actions('http://js-use-enc')):
            self.skipTest("external resource not available")
        """Test generating a suggested parent item for a given item"""
        new_wkt = 'clean up the house'
        self.run_walker('gen_rand_life', {})
        self.run_walker('get_gen_day', {})
        data = self.run_walker('get_latest_day', {'show_report': 1})['report']
        w_id = data[0][1]['jid']
        data = self.run_walker(
            'get_suggested_parent', {'new_wkt_name': new_wkt}, prime=w_id)
        self.assertTrue(len(data['report']) > 0)
        self.assertTrue(0 < data['report'][-1][1] < 1)

    def test_due_soon(self):
        """Test generating a list of suggested focus items for a given day"""
        self.run_walker('gen_rand_life', {})
        self.run_walker('get_gen_day', {})
        data = self.run_walker('get_latest_day', {'show_report': 1})
        w_id = data['report'][0][1]['jid']

        today_date = datetime.datetime.today()
        due_item = False
        for wkt in data['report']:
            if wkt[1]['name'] == 'workette':
                if random.choice([0, 1]) or not due_item:
                    due_item = True
                    due = random.randint(-3, 3)
                    due_date = today_date + datetime.timedelta(days=due)
                    due_date = due_date.replace(
                        hour=0, minute=0, second=0, microsecond=0)
                    wkt[1]['context']['date'] = due_date.isoformat()
                    self.graph_node_set(wkt[1]['jid'], wkt[1]['context'])

        data = self.run_walker(
            'get_due_soon', {'soon': 4, 'show_report': 1}, prime=w_id)
        self.assertTrue(len(data['report']) > 0)

    def test_days_in_backlog(self):
        """Test getting the number of days an item has been in the backlog"""
        self.run_walker('gen_rand_life', {})
        self.run_walker('get_gen_day', {})
        data = self.run_walker('get_latest_day', {'show_report': 1})
        w_id = data['report'][0][1]['jid']
        workettes = self.run_walker('get_workettes', {}, prime=w_id)['report']
        for wkt in workettes:
            res = self.run_walker(
                'days_in_backlog', {'show_report': 1}, prime=wkt['jid'])
            self.assertIs(type(res['report'][0]), int)

    def test_get_long_active_items(self):
        """Test getting items been in backlog for long"""
        self.run_walker('gen_rand_life', {})
        self.run_walker('get_gen_day', {})
        data = self.run_walker('get_latest_day', {'show_report': 1})
        w_id = data['report'][0][1]['jid']
        result = self.run_walker(
            'get_long_active_items',
            {'show_report': 1, 'long_days': 1}, prime=w_id)
        self.assertTrue(len(result) > 0)

    def test_get_suggested_focus(self):
        self.run_walker('gen_rand_life', {})
        self.run_walker('get_gen_day', {})
        data = self.run_walker('get_latest_day', {'show_report': 1})
        w_id = data['report'][0][1]['jid']
        result = self.run_walker(
            'get_suggested_focus',
            {'max_items': 5, 'long_days': 1}, prime=w_id)
        self.assertTrue(len(result) > 0)



from django.contrib import admin
from django.contrib.auth.admin import UserAdmin as BaseUserAdmin
from django.utils.translation import gettext as _

from jaseci_serv.base import models


class UserAdmin(BaseUserAdmin):
    """
    Customized user listing for admin page
    """
    ordering = ['time_created']
    list_display = ['email', 'name', 'time_created']
    fieldsets = (
        (None, {'fields': ('email', 'password')}),
        (_('Personal Info'), {'fields': ('name',)}),
        (
            _('Permissions'),
            {
                'fields': (
                    'is_active',
                    'is_activated',
                    'is_staff',
                    'is_admin',
                    'is_superuser',
                )
            }
        ),
        (_('Important dates'), {'fields': ('last_login',)}),
    )
    add_fieldsets = (
        (None, {
            'classes': ('wide',),
            'fields': ('email', 'password1', 'password2')
        }),
    )


admin.site.register(models.User, UserAdmin)


class JaseciObjectAdmin (admin.ModelAdmin):
    ordering = ['j_timestamp']
    list_display = ('jid', 'name', 'j_type', 'j_parent',
                    'j_master', 'j_timestamp')
    search_fields = ['jid', 'name', 'j_type', 'j_parent',
                     'j_master', 'j_timestamp', ]


admin.site.register(models.JaseciObject, JaseciObjectAdmin)


class GlobalVarsAdmin (admin.ModelAdmin):
    ordering = ['name']
    list_display = ('name', 'value')
    search_fields = ['name', 'value']


admin.site.register(models.GlobalVars, GlobalVarsAdmin)



"""
This module includes code related to hooking Jaseci's django models to the
core engine.

FIX: Serious permissions work needed
"""
from django.core.exceptions import ObjectDoesNotExist

from jaseci.utils import utils
from jaseci.utils.utils import logger
import jaseci as core_mod
from jaseci.utils.mem_hook import mem_hook, json_str_to_jsci_dict
from jaseci_serv. jaseci_serv.settings import REDIS_HOST
from redis import Redis
import uuid
import json


def find_class_and_import(j_type, core_mod):
    if(j_type == 'master'):
        from jaseci_serv.base.models import master
        return master
    elif(j_type == 'super_master'):
        from jaseci_serv.base.models import super_master
        return super_master
    else:
        return utils.find_class_and_import(j_type, core_mod)


class redis_cache():
    """Redis sideline cache"""

    def __init__(self, red):
        self.red = red

    def get(self, urn):
        try:
            return self.red.get(urn)
        except Exception:
            return None

    def set(self, urn, value):
        try:
            return self.red.set(urn, value)
        except TypeError:
            logger.error(
                f"Item {value} is not JSON serializable for redis store!")
            logger.error(f"Item details: {value.serialize()}",  exc_info=True)
        except Exception:
            return None

    def delete(self, urn):
        try:
            return self.red.delete(urn)
        except Exception:
            return None


class orm_hook(mem_hook):
    """
    Hooks Django ORM database for Jaseci objects to Jaseci's core engine.
    """

    def __init__(self, objects, globs,
                 red=Redis(host=REDIS_HOST, decode_responses=True)):
        self.objects = objects
        self.globs = globs
        self.red = redis_cache(red)
        self.skip_redis_update = False
        self.save_obj_list = set()
        self.save_glob_dict = {}
        super().__init__()

    def get_obj_from_store(self, item_id):
        loaded_obj = self.red.get(item_id.urn)
        if (loaded_obj):
            jdict = json.loads(loaded_obj)
            j_type = jdict['j_type']
            j_master = jdict['j_master']
            class_for_type = \
                find_class_and_import(j_type, core_mod)
            ret_obj = class_for_type(
                h=self, m_id=j_master, auto_save=False)
            ret_obj.json_load(loaded_obj)

            return ret_obj
        else:
            try:
                loaded_obj = self.objects.get(
                    jid=item_id)
            except ObjectDoesNotExist:
                logger.error(
                    str(f"Object {item_id} does not exist in Django ORM!"),
                    exc_info=True
                )
                return None

            class_for_type = \
                find_class_and_import(loaded_obj.j_type, core_mod)
            ret_obj = class_for_type(
                h=self, m_id=loaded_obj.j_master.urn, auto_save=False)
            utils.map_assignment_of_matching_fields(ret_obj, loaded_obj)
            assert(uuid.UUID(ret_obj.jid) == loaded_obj.jid)

            # Unwind jsci_payload for fields beyond element object
            obj_fields = json_str_to_jsci_dict(loaded_obj.jsci_obj, ret_obj)
            for i in obj_fields.keys():
                setattr(ret_obj, i, obj_fields[i])

            self.red.set(ret_obj.id.urn, ret_obj.json(detailed=True))
            return ret_obj

    def has_obj_in_store(self, item_id):
        """
        Checks for object existance in store
        """
        if (self.red.get(item_id.urn)):
            return True
        return self.objects.filter(jid=item_id).count()

    def save_obj_to_store(self, item):
        # import traceback as tb; tb.print_stack();  # noqa
        self.save_obj_list.add(item)

    def commit_obj_to_redis(self, item):
        self.red.set(item.id.urn, item.json(detailed=True))

    def commit_obj(self, item):
        item_from_db, created = self.objects.get_or_create(jid=item.id
                                                           )
        utils.map_assignment_of_matching_fields(item_from_db, item)
        item_from_db.jsci_obj = item.jsci_payload()
        item_from_db.save()

    def destroy_obj_from_store(self, item):
        try:
            self.objects.get(jid=item.id).delete()
        except ObjectDoesNotExist:
            # NOTE: Should look at this at some point
            # logger.error("Object does not exists so delete aborted!")
            pass
        self.red.delete(item.id.urn)
        if(item in self.save_obj_list):
            self.save_obj_list.remove(item)

    def get_glob_from_store(self, name):
        """
        Get global config from externally hooked general store by name
        """
        loaded_val = self.red.get(name)
        if (loaded_val):
            return loaded_val
        else:
            try:
                loaded_val = self.globs.get(name=name).value
            except ObjectDoesNotExist:
                logger.error(
                    str(f"Global {name} does not exist in Django ORM!"),
                    exc_info=True
                )
                return None

            self.red.set(name, loaded_val)
            return loaded_val

    def has_glob_in_store(self, name):
        """
        Checks for global config existance in store
        """
        if (self.red.get(name)):
            return True
        return self.globs.filter(name=name).count()

    def save_glob_to_store(self, name, value):
        """Save global config to externally hooked general store"""
        self.save_glob_dict[name] = value

    def list_glob_from_store(self):
        """Get list of global config to externally hooked general store"""
        return [entry['name'] for entry in self.globs.values('name')]

    def destroy_glob_from_store(self, name):
        """Destroy global config to externally hooked general store"""
        try:
            self.globs.get(name=name).delete()
        except ObjectDoesNotExist:
            pass
        self.red.delete(name)
        if(name in self.save_glob_dict.keys()):
            del self.save_glob_dict[name]

    def commit_glob(self, name, value):
        try:
            self.red.set(name, value)
        except Exception as e:
            logger.error(
                str(f"Couldn't save {name} to redis! {e}"),
                exc_info=True
            )
        item_from_db, created = self.globs.get_or_create(name=name)
        item_from_db.value = value
        item_from_db.save()

    def commit(self):
        """Write through all saves to store"""
        # dist = {}
        # for i in self.save_obj_list:
        #     if (type(i).__name__ in dist.keys()):
        #         dist[type(i).__name__] += 1
        #     else:
        #         dist[type(i).__name__] = 1
        # print(dist)
        for i in self.save_obj_list:
            if(not self.skip_redis_update):
                self.commit_obj_to_redis(i)
            else:
                self.skip_redis_update = False
            self.commit_obj(i)
        self.save_obj_list = set()
        for i in self.save_glob_dict.keys():
            self.commit_glob(name=i, value=self.save_glob_dict[i])
        self.save_glob_dict = {}
